#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.

class Lazy:
    def __init__(self, field, f, deps):
        self.field = field
        self.f = f
        self.deps = deps

    def __call__(self, *args, **kwargs):
        for k, v in self.deps.items():
            if isinstance(v, Lazy):
                self.deps[k] = v()
        return self.f(**self.deps)[self.field]

    def __repr__(self):
        dic = {}
        for k, v in self.deps.items():
            dic[k] = v if isinstance(v, Lazy) else ""
        return f"→({' '.join([f'{k}{v}' for k,v in dic.items()])})"

